@@ -75,7 +75,8 @@ enum {
 	PROP_INSERT_SPACES,
 	PROP_SHOW_MARGIN,
 	PROP_MARGIN,
-	PROP_SMART_HOME_END
+	PROP_SMART_HOME_END,
+	PROP_HIGHLIGHT_CURRENT_LINE
 };
 
 
@@ -87,6 +88,7 @@ struct _GtkSourceViewPrivate
 	gboolean	 auto_indent;
 	gboolean	 insert_spaces;
 	gboolean	 show_margin;
+	gboolean	 highlight_current_line;
 	guint		 margin;
 	gint             cached_margin_width;
 	gboolean	 smart_home_end;
@@ -171,7 +173,6 @@ static void	gtk_source_view_get_property		(GObject           *object,
 static void     gtk_source_view_style_set               (GtkWidget         *widget,
 							 GtkStyle          *previous_style);
 
-static void	gtk_source_view_grab_focus		(GtkWidget         *widget);
 
 /* Private functions. */
 static void
@@ -193,7 +194,6 @@ gtk_source_view_class_init (GtkSourceViewClass *klass)
 	
 	widget_class->expose_event = gtk_source_view_expose;
 	widget_class->style_set = gtk_source_view_style_set;
-	widget_class->grab_focus = gtk_source_view_grab_focus;
 
 	textview_class->populate_popup = gtk_source_view_populate_popup;
 	textview_class->move_cursor = gtk_source_view_move_cursor;
@@ -249,7 +249,7 @@ gtk_source_view_class_init (GtkSourceViewClass *klass)
 							       _("Whether to display the right margin"),
 							       FALSE,
 							       G_PARAM_READWRITE));
-
+	
 	g_object_class_install_property (object_class,
 					 PROP_MARGIN,
 					 g_param_spec_uint ("margin",
@@ -264,10 +264,20 @@ gtk_source_view_class_init (GtkSourceViewClass *klass)
 					 PROP_SMART_HOME_END,
 					 g_param_spec_boolean ("smart_home_end",
 							       _("Use smart home/end"),
-							       _("HOME and END keys move to first/last characters on line first before going to the start/end of the line"),
+							       _("HOME and END keys move to first/last "
+								 "characters on line first before going "
+								 "to the start/end of the line"),
 							       TRUE,
 							       G_PARAM_READWRITE));
 
+	g_object_class_install_property (object_class,
+					 PROP_HIGHLIGHT_CURRENT_LINE,
+					 g_param_spec_boolean ("highlight_current_line",
+							       _("Highlight current line"),
+							       _("Whether to highlight the current line"),
+							       FALSE,
+							       G_PARAM_READWRITE));
+
 	signals [UNDO] =
 		g_signal_new ("undo",
 			      G_TYPE_FROM_CLASS (klass),
@@ -299,6 +309,10 @@ gtk_source_view_class_init (GtkSourceViewClass *klass)
 				      GDK_z,
 				      GDK_CONTROL_MASK | GDK_SHIFT_MASK,
 				      "redo", 0);
+	gtk_binding_entry_add_signal (binding_set,
+				      GDK_F14,
+				      0,
+				      "undo", 0);
 }
 
 static void 
@@ -355,6 +369,12 @@ gtk_source_view_set_property (GObject      *object,
 			gtk_source_view_set_smart_home_end (view,
 							    g_value_get_boolean (value));
 			break;
+
+		case PROP_HIGHLIGHT_CURRENT_LINE:
+			gtk_source_view_set_highlight_current_line (view,
+								    g_value_get_boolean (value));
+			break;
+
 		
 		default:
 			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -421,6 +441,10 @@ gtk_source_view_get_property (GObject    *object,
 					     gtk_source_view_get_smart_home_end (view));
 			break;
 			
+		case PROP_HIGHLIGHT_CURRENT_LINE:
+			g_value_set_boolean (value,
+					     gtk_source_view_get_highlight_current_line (view));
+
 		default:
 			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 			break;
@@ -970,6 +994,8 @@ gtk_source_view_paint_margin (GtkSourceView *view,
 	gint margin_width;
 	gint text_width, x_pixmap;
 	gint i;
+	GtkTextIter cur;
+	gint cur_line;
 
 	text_view = GTK_TEXT_VIEW (view);
 
@@ -1091,10 +1117,17 @@ gtk_source_view_paint_margin (GtkSourceView *view,
 		marker_line = gtk_source_marker_get_line (
 			GTK_SOURCE_MARKER (current_marker->data));
 	
+	
+	gtk_text_buffer_get_iter_at_mark (text_view->buffer, 
+					  &cur, 
+					  gtk_text_buffer_get_insert (text_view->buffer));
+
+	cur_line = gtk_text_iter_get_line (&cur) + 1;
+	
 	while (i < count) 
 	{
 		gint pos;
-
+		
 		gtk_text_view_buffer_to_window_coords (text_view,
 						       GTK_TEXT_WINDOW_LEFT,
 						       0,
@@ -1104,10 +1137,24 @@ gtk_source_view_paint_margin (GtkSourceView *view,
 
 		if (view->priv->show_line_numbers) 
 		{
-			g_snprintf (str, sizeof (str),
-				    "%d", g_array_index (numbers, gint, i) + 1);
+			gint line_to_paint = g_array_index (numbers, gint, i) + 1;
+			
+			if (line_to_paint == cur_line)
+			{
+				gchar *markup;
+				markup = g_strdup_printf ("<b>%d</b>", line_to_paint);
+				
+				pango_layout_set_markup (layout, markup, -1);
 
-			pango_layout_set_text (layout, str, -1);
+				g_free (markup);
+			}
+			else
+			{
+				g_snprintf (str, sizeof (str),
+					    "%d", line_to_paint);
+
+				pango_layout_set_markup (layout, str, -1);
+			}
 
 			gtk_paint_layout (GTK_WIDGET (view)->style,
 					  win,
@@ -1213,6 +1260,54 @@ gtk_source_view_expose (GtkWidget      *widget,
 				gdk_window_invalidate_rect (w, NULL, FALSE);
 		}
 
+		if (view->priv->highlight_current_line && 
+		    (event->window == gtk_text_view_get_window (text_view, GTK_TEXT_WINDOW_TEXT)))
+		{
+			GdkRectangle visible_rect;
+			GdkRectangle redraw_rect;
+			GtkTextIter cur;
+			gint y;
+			gint height;
+			gint win_y;
+
+			gtk_text_buffer_get_iter_at_mark (text_view->buffer, 
+							  &cur, 
+							  gtk_text_buffer_get_insert (text_view->buffer));
+
+			gtk_text_view_get_line_yrange (text_view, &cur, &y, &height);
+							
+			gtk_text_view_get_visible_rect (text_view, &visible_rect);
+			
+			gtk_text_view_buffer_to_window_coords (text_view,
+						       GTK_TEXT_WINDOW_TEXT,
+						       visible_rect.x,
+						       visible_rect.y,
+						       &redraw_rect.x,
+						       &redraw_rect.y);
+
+			gtk_text_view_buffer_to_window_coords (text_view,
+						       GTK_TEXT_WINDOW_TEXT,
+						       0,
+						       y,
+						       NULL,
+						       &win_y);
+
+			redraw_rect.width = visible_rect.width;
+			redraw_rect.height = visible_rect.height;
+
+			gtk_paint_flat_box (widget->style, 
+					    event->window, 
+					    GTK_WIDGET_STATE (widget),
+					    GTK_SHADOW_NONE,
+					    &redraw_rect, 
+					    widget,
+					    "current_line", 
+					    MAX (1, gtk_text_view_get_left_margin (text_view) - 1),
+					    win_y,
+					    redraw_rect.width,
+					    height);
+		}
+
 		if (view->priv->show_margin && 
 		    (event->window == gtk_text_view_get_window (text_view, GTK_TEXT_WINDOW_TEXT)))
 		{
@@ -1248,6 +1343,7 @@ gtk_source_view_expose (GtkWidget      *widget,
 					 gtk_text_view_get_left_margin (text_view));
 		}
 
+		
 		if (GTK_WIDGET_CLASS (parent_class)->expose_event)
 			event_handled = 
 				(* GTK_WIDGET_CLASS (parent_class)->expose_event)
@@ -1300,6 +1396,15 @@ calculate_real_tab_width (GtkSourceView *view, guint tab_size, gchar c)
  * Public interface 
  * ---------------------------------------------------------------------- */
 
+/**
+ * gtk_source_view_new:
+ *
+ * Creates a new #GtkSourceView. An empty default buffer will be
+ * created for you. If you want to specify your own buffer, consider
+ * gtk_source_view_new_with_buffer().
+ *
+ * Return value: a new #GtkSourceView
+ **/
 GtkWidget *
 gtk_source_view_new ()
 {
@@ -1312,11 +1417,22 @@ gtk_source_view_new ()
 	return widget;
 }
 
+/**
+ * gtk_source_view_new_with_buffer:
+ * @buffer: a #GtkSourceBuffer
+ *
+ * Creates a new #GtkSourceView widget displaying the buffer
+ * @buffer. One buffer can be shared among many widgets.
+ *
+ * Return value: a new #GtkTextView.
+ **/
 GtkWidget *
 gtk_source_view_new_with_buffer (GtkSourceBuffer *buffer)
 {
 	GtkWidget *view;
 
+	g_return_val_if_fail (buffer != NULL && GTK_IS_SOURCE_BUFFER (buffer), NULL);
+	
 	view = g_object_new (GTK_TYPE_SOURCE_VIEW, NULL);
 	gtk_text_view_set_buffer (GTK_TEXT_VIEW (view), GTK_TEXT_BUFFER (buffer));
 
@@ -1349,6 +1465,14 @@ gtk_source_view_get_type (void)
 	return our_type;
 }
 
+/**
+ * gtk_source_view_get_show_line_numbers:
+ * @view: a #GtkSourceView
+ *
+ * Returns whether line numbers are displayed beside the text.
+ *
+ * Return value: TRUE if the line numbers are displayed.
+ **/
 gboolean
 gtk_source_view_get_show_line_numbers (GtkSourceView *view)
 {
@@ -1358,16 +1482,24 @@ gtk_source_view_get_show_line_numbers (GtkSourceView *view)
 	return view->priv->show_line_numbers;
 }
 
+/**
+ * gtk_source_view_set_show_line_numbers:
+ * @view: a #GtkSourceView
+ * @show: whether line numbers should be displayed
+ *
+ * If TRUE line numbers will be displayed beside the text.
+ *
+ **/
 void
 gtk_source_view_set_show_line_numbers (GtkSourceView *view,
-				       gboolean       visible)
+				       gboolean       show)
 {
 	g_return_if_fail (view != NULL);
 	g_return_if_fail (GTK_IS_SOURCE_VIEW (view));
 
-	visible = (visible != FALSE);
+	show = (show != FALSE);
 
-	if (visible) 
+	if (show) 
 	{
 		if (!view->priv->show_line_numbers) 
 		{
@@ -1381,16 +1513,16 @@ gtk_source_view_set_show_line_numbers (GtkSourceView *view,
 			else
 				gtk_widget_queue_draw (GTK_WIDGET (view));
 
-			view->priv->show_line_numbers = visible;
+			view->priv->show_line_numbers = show;
 
 			g_object_notify (G_OBJECT (view), "show_line_numbers");
 		}
-	} 
+	}
 	else 
 	{
 		if (view->priv->show_line_numbers) 
 		{
-			view->priv->show_line_numbers = visible;
+			view->priv->show_line_numbers = show;
 
 			/* force expose event, which will adjust margin. */
 			gtk_widget_queue_draw (GTK_WIDGET (view));
@@ -1400,6 +1532,14 @@ gtk_source_view_set_show_line_numbers (GtkSourceView *view,
 	}
 }
 
+/**
+ * gtk_source_view_get_show_line_markers:
+ * @view: a #GtkSourceView
+ *
+ * Returns whether line markers are displayed beside the text.
+ *
+ * Return value: TRUE if the line markers are displayed.
+ **/
 gboolean
 gtk_source_view_get_show_line_markers (GtkSourceView *view)
 {
@@ -1409,16 +1549,24 @@ gtk_source_view_get_show_line_markers (GtkSourceView *view)
 	return view->priv->show_line_markers;
 }
 
+/**
+ * gtk_source_view_set_show_line_markers:
+ * @view: a #GtkSourceView
+ * @show: whether line markers should be displayed
+ *
+ * If TRUE line markers will be displayed beside the text.
+ *
+ **/
 void
 gtk_source_view_set_show_line_markers (GtkSourceView *view,
-				       gboolean       visible)
+				       gboolean       show)
 {
 	g_return_if_fail (view != NULL);
 	g_return_if_fail (GTK_IS_SOURCE_VIEW (view));
 
-	visible = (visible != FALSE);
+	show = (show != FALSE);
 
-	if (visible) 
+	if (show) 
 	{
 		if (!view->priv->show_line_markers) 
 		{
@@ -1432,7 +1580,7 @@ gtk_source_view_set_show_line_markers (GtkSourceView *view,
 			else
 				gtk_widget_queue_draw (GTK_WIDGET (view));
 
-			view->priv->show_line_markers = visible;
+			view->priv->show_line_markers = show;
 
 			g_object_notify (G_OBJECT (view), "show_line_markers");
 		}
@@ -1441,7 +1589,7 @@ gtk_source_view_set_show_line_markers (GtkSourceView *view,
 	{
 		if (view->priv->show_line_markers) 
 		{
-			view->priv->show_line_markers = visible;
+			view->priv->show_line_markers = show;
 
 			/* force expose event, which will adjust margin. */
 			gtk_widget_queue_draw (GTK_WIDGET (view));
@@ -1451,6 +1599,14 @@ gtk_source_view_set_show_line_markers (GtkSourceView *view,
 	}
 }
 
+/**
+ * gtk_source_view_get_tabs_width:
+ * @view: a #GtkSourceView
+ *
+ * Returns the width of tabulation in characters.
+ *
+ * Return value: width of tab.
+ **/
 guint
 gtk_source_view_get_tabs_width (GtkSourceView *view)
 {
@@ -1482,6 +1638,14 @@ set_tab_stops_internal (GtkSourceView *view)
 	return TRUE;
 }
 
+/**
+ * gtk_source_view_set_tabs_width:
+ * @view: a #GtkSourceView
+ * @width: width of tab in characters
+ *
+ * Sets the width of tabulation in characters.
+ *
+ **/
 void
 gtk_source_view_set_tabs_width (GtkSourceView *view,
 				guint          width)
@@ -1510,6 +1674,14 @@ gtk_source_view_set_tabs_width (GtkSourceView *view,
 	}
 }
 
+/**
+ * gtk_source_view_set_marker_pixbuf:
+ * @view: a #GtkSourceView
+ * @marker_type: 
+ * @pixbuf:
+ *
+ *
+ **/
 void
 gtk_source_view_set_marker_pixbuf (GtkSourceView *view,
 				   const gchar   *marker_type,
@@ -1552,6 +1724,13 @@ gtk_source_view_set_marker_pixbuf (GtkSourceView *view,
 	}
 }
 
+/**
+ * gtk_source_view_get_marker_pixbuf:
+ * @view: a #GtkSourceView
+ * @marker_type: 
+ *
+ * Return value:
+ **/
 GdkPixbuf * 
 gtk_source_view_get_marker_pixbuf (GtkSourceView *view,
 				   const gchar   *marker_type)
@@ -1584,7 +1763,7 @@ compute_indentation (GtkSourceView *view, gint line)
 	end = start;
 
 	ch = gtk_text_iter_get_char (&end);
-	while (g_unichar_isspace (ch) && ch != '\n')
+	while (g_unichar_isspace (ch) && ch != '\n' && ch != '\r')
 	{
 		if (!gtk_text_iter_forward_char (&end))
 			break;
@@ -1615,8 +1794,7 @@ key_press_cb (GtkWidget *widget, GdkEventKey *event, gpointer data)
 	mark = gtk_text_buffer_get_insert (buf);
 	gtk_text_buffer_get_iter_at_mark (buf, &cur, mark);
 	
-	if ((key == GDK_Return) && gtk_text_iter_ends_line (&cur) && 
-	    view->priv->auto_indent)
+	if ((key == GDK_Return) && view->priv->auto_indent)
 	{
 		/* Auto-indent means that when you press ENTER at the end of a
 		 * line, the new line is automatically indented at the same
@@ -1672,6 +1850,14 @@ key_press_cb (GtkWidget *widget, GdkEventKey *event, gpointer data)
 	return FALSE;
 }
 
+/**
+ * gtk_source_view_get_auto_indent:
+ * @view: a #GtkSourceView
+ *
+ * Returns whether auto indentation of text is enabled.
+ *
+ * Return value: TRUE if auto indentation is enabled.
+ **/
 gboolean
 gtk_source_view_get_auto_indent (GtkSourceView *view)
 {
@@ -1680,6 +1866,14 @@ gtk_source_view_get_auto_indent (GtkSourceView *view)
 	return view->priv->auto_indent;
 }
 
+/**
+ * gtk_source_view_set_auto_indent:
+ * @view: a #GtkSourceView
+ * @enable: whether to enable auto indentation
+ *
+ * If TRUE auto indentation of text is enabled.
+ *
+ **/
 void
 gtk_source_view_set_auto_indent (GtkSourceView *view, gboolean enable)
 {
@@ -1695,6 +1889,15 @@ gtk_source_view_set_auto_indent (GtkSourceView *view, gboolean enable)
 	g_object_notify (G_OBJECT (view), "auto_indent");
 }
 
+/**
+ * gtk_source_view_get_insert_spaces_instead_of_tabs:
+ * @view: a #GtkSourceView
+ *
+ * Returns whether when inserting a tabulator character it should
+ * be replaced by a group of space characters.
+ *
+ * Return value: TRUE if spaces are inserted instead of tabs.
+ **/
 gboolean
 gtk_source_view_get_insert_spaces_instead_of_tabs (GtkSourceView *view)
 {
@@ -1703,6 +1906,15 @@ gtk_source_view_get_insert_spaces_instead_of_tabs (GtkSourceView *view)
 	return view->priv->insert_spaces;
 }
 
+/**
+ * gtk_source_view_set_insert_spaces_instead_of_tabs:
+ * @view: a #GtkSourceView
+ * @enable: whether to insert spaces instead of tabs
+ *
+ * If TRUE any tabulator character inserted is replaced by a group
+ * of space characters.
+ *
+ **/
 void
 gtk_source_view_set_insert_spaces_instead_of_tabs (GtkSourceView *view, gboolean enable)
 {
@@ -1783,6 +1995,29 @@ view_dnd_drop (GtkTextView *view,
 	}
 }
 
+/**
+ * gtk_source_view_get_show_margin:
+ * @view: a #GtkSourceView
+ *
+ * Returns whether a margin is displayed
+ *
+ * Return value: TRUE if the margin is showed
+ **/
+gboolean 
+gtk_source_view_get_show_margin (GtkSourceView *view)
+{
+	g_return_val_if_fail (GTK_IS_SOURCE_VIEW (view), FALSE);
+
+	return view->priv->show_margin;
+}
+
+/**
+ * gtk_source_view_set_show_margin:
+ * @view: a #GtkSourceView
+ * @show: whether to show a margin
+ *
+ * If TRUE a margin is displayed
+ **/
 void 
 gtk_source_view_set_show_margin (GtkSourceView *view, gboolean show)
 {
@@ -1800,14 +2035,70 @@ gtk_source_view_set_show_margin (GtkSourceView *view, gboolean show)
 	g_object_notify (G_OBJECT (view), "show_margin");
 }
 
+/**
+ * gtk_source_view_get_highlight_current_line:
+ * @view: a #GtkSourceView
+ *
+ * Returns whether the current line is highlighted
+ *
+ * Return value: TRUE if the current line is highlighted
+ **/
 gboolean 
-gtk_source_view_get_show_margin (GtkSourceView *view)
+gtk_source_view_get_highlight_current_line (GtkSourceView *view)
 {
 	g_return_val_if_fail (GTK_IS_SOURCE_VIEW (view), FALSE);
 
-	return view->priv->show_margin;
+	return view->priv->highlight_current_line;
 }
 
+/**
+ * gtk_source_view_set_highlight_current_line:
+ * @view: a #GtkSourceView
+ * @show: whether to highlight the current line
+ *
+ * If TRUE the current line is highlighted
+ **/
+void 
+gtk_source_view_set_highlight_current_line (GtkSourceView *view, gboolean hl)
+{
+	g_return_if_fail (GTK_IS_SOURCE_VIEW (view));
+
+	hl = (hl != FALSE);
+
+	if (view->priv->highlight_current_line == hl)
+		return;
+
+	view->priv->highlight_current_line = hl;
+
+	gtk_widget_queue_draw (GTK_WIDGET (view));
+
+	g_object_notify (G_OBJECT (view), "highlight_current_line");
+}
+
+/**
+ * gtk_source_view_get_margin:
+ * @view: a #GtkSourceView
+ *
+ *
+ *
+ * Return value:
+ **/
+guint
+gtk_source_view_get_margin  (GtkSourceView *view)
+{
+	g_return_val_if_fail (GTK_IS_SOURCE_VIEW (view), DEFAULT_MARGIN);
+
+	return view->priv->margin;
+
+}
+
+/**
+ * gtk_source_view_set_margin:
+ * @view: a #GtkSourceView
+ *
+ *
+ *
+ **/
 void 
 gtk_source_view_set_margin (GtkSourceView *view, guint margin)
 {
@@ -1825,16 +2116,16 @@ gtk_source_view_set_margin (GtkSourceView *view, guint margin)
 
 	g_object_notify (G_OBJECT (view), "margin");
 }
-		
-guint
-gtk_source_view_get_margin  (GtkSourceView *view)
-{
-	g_return_val_if_fail (GTK_IS_SOURCE_VIEW (view), DEFAULT_MARGIN);
-
-	return view->priv->margin;
-
-}
 
+/**
+ * gtk_source_view_set_smart_home_end:
+ * @view: a #GtkSourceView
+ * @enable: whether to enable smart behavior for HOME and END keys
+ *
+ * If TRUE HOME and END keys will move to the first/last non-space
+ * character of the line before moving to the start/end.
+ *
+ **/
 void
 gtk_source_view_set_smart_home_end (GtkSourceView *view, gboolean enable)
 {
@@ -1850,6 +2141,15 @@ gtk_source_view_set_smart_home_end (GtkSourceView *view, gboolean enable)
 	g_object_notify (G_OBJECT (view), "smart_home_end");
 }
 
+/**
+ * gtk_source_view_get_smart_home_end:
+ * @view: a #GtkSourceView
+ *
+ * Returns whether HOME and END keys will move to the first/last non-space
+ * character of the line before moving to the start/end.
+ *
+ * Return value: TRUE if smart behavior for HOME and END keys is enabled
+ **/
 gboolean
 gtk_source_view_get_smart_home_end (GtkSourceView *view)
 {
@@ -1858,6 +2158,13 @@ gtk_source_view_get_smart_home_end (GtkSourceView *view)
 	return view->priv->smart_home_end;
 }
 
+/**
+ * gtk_source_view_style_set:
+ * @widget: a #GtkSourceView
+ * @previous_style:
+ *
+ *
+ **/
 static void 
 gtk_source_view_style_set (GtkWidget *widget, GtkStyle *previous_style)
 {
@@ -1883,21 +2190,3 @@ gtk_source_view_style_set (GtkWidget *widget, GtkStyle *previous_style)
 	}
 }
 
-/* This is a workaround for bug #81893.
- * We skip the parent class' grab_focus method in order to avoid the weird 
- * scroll on focus behavior of GtkTextView.
- * The GtkTextView grab_focus method first calls the parent class' 
- * grab_focus method and then scrolls to cursor if disable_scroll_on_focus is
- * FALSE.
- * We will remove this workaround when bug #81893 will be fixed.
- */
-static void
-gtk_source_view_grab_focus (GtkWidget *widget)
-{
-  GtkTextView *text_view;
-
-  text_view = GTK_TEXT_VIEW (widget);
-  
-  GTK_WIDGET_CLASS (g_type_class_peek_parent (parent_class))->grab_focus (widget);
-}
-
